// console variables

#include "cvars.h"
#include "error.h"
#include "wnd_volume.h"


typedef std::vector<cvar_t> vec_cvar_t;
vec_cvar_t vec_cvars;


// it is better to add all variables right through here; otherwise later may appear a mess, when some vars will appear registered, and some aren't yet
void Cvars_Init ()
{
	CVARS_ADD_BOOL(wnd_volume_show);
	CVARS_ADD_BOOL(wnd_volume_discoloured);
	CVARS_ADD_BYTE(wnd_volume_opacity);
	CVARS_ADD_UINT(wnd_volume_timer_elapse_ms);
}

void Cvars_Quit ()
{
}

void Cvars_Add (cvar_t& cvar)
{
	vec_cvars.push_back(cvar);
}

void Cvars_Add_Bool (const char* name, bool* var_ptr)
{
	cvar_t cvar;

	Assert(name);
	Assert(var_ptr);
	cvar.name = name;
	cvar.type = CVAR_TYPE__BOOL;
	cvar.var_ptr = var_ptr;
	Cvars_Add (cvar);
}

void Cvars_Add_Byte (const char* name, byte* var_ptr)
{
	cvar_t cvar;

	Assert(name);
	Assert(var_ptr);
	cvar.name = name;
	cvar.type = CVAR_TYPE__BYTE;
	cvar.var_ptr = var_ptr;
	Cvars_Add (cvar);
}

void Cvars_Add_Uint (const char* name, uint* var_ptr)
{
	cvar_t cvar;

	Assert(name);
	Assert(var_ptr);
	cvar.name = name;
	cvar.type = CVAR_TYPE__UINT;
	cvar.var_ptr = var_ptr;
	Cvars_Add (cvar);
}

static bool Cvars_ParseCommand (cvar_t& cvar, vec_str_t& vec_str_args)
{
	uint args_num;
	uint args_num_necessary;
	bool val_bool;
//	byte val_byte;
	uint val_uint;
	stringstream strm;
	const char* type_name;
	string* arg0;
	const char* arg0_str;

// enable hexadecimal input
//// hehe, "0x" prefix already is in support
//	strm << hex;
// *'dec' disables or overrides 'hex'
//	strm << dec;
//	strm.setf(ios::dec);
	arg0 = &vec_str_args[0];
	arg0_str = arg0->c_str();
// so do a trick :  define by self, if it is "0x"
//	if (! strnicmp (arg0->c_str(), "0x"))
	if (arg0_str[0] == '0' && (arg0_str[1] == 'x' || arg0_str[1] == 'X'))
		strm << hex;
	args_num = vec_str_args.size();
	switch (cvar.type)
	{
	case CVAR_TYPE__BOOL:
		type_name = "bool";
		args_num_necessary = 1;
		if (args_num != args_num_necessary)
			goto wrong_args_num;
		if (! cvar.var_ptr)
			goto no_cvar_ptr;
		strm << vec_str_args[0];
		strm >> val_bool;
		if (strm.fail() || ! strm.eof())
//			goto error_arg_numeric;
//			Error ("can't read arg \"%s\" as a bool", vec_str_args[0].c_str());
			goto unreadable_arg;
		*(bool*)cvar.var_ptr = val_bool;
		break;

	case CVAR_TYPE__BYTE:
		type_name = "byte";
		args_num_necessary = 1;
		if (args_num != args_num_necessary)
			goto wrong_args_num;
		if (! cvar.var_ptr)
			goto no_cvar_ptr;
		strm << vec_str_args[0];
//		strm >> val_byte;
		strm >> val_uint;
		if (strm.fail() || ! strm.eof())
			goto unreadable_arg;
		if (val_uint >= 0x100)
			goto unreadable_arg;
//		*(byte*)cvar.var_ptr = val_byte;
		*(byte*)cvar.var_ptr = (byte)val_uint;
		break;

	case CVAR_TYPE__UINT:
		type_name = "uint";
		args_num_necessary = 1;
		if (args_num != args_num_necessary)
			goto wrong_args_num;
		if (! cvar.var_ptr)
			goto no_cvar_ptr;
		strm << vec_str_args[0];
		strm >> val_uint;
		if (strm.fail() || ! strm.eof())
//			Error ("can't read arg \"%s\" as an uint", vec_str_args[0].c_str());
			goto unreadable_arg;
		*(uint*)cvar.var_ptr = val_uint;
		break;

	default:
		Error ("Cvars_ParseCommand () :  cvar type %u is unknown", cvar.type);
	}

	return 1;

wrong_args_num:
	args_num, args_num_necessary;
	return 0;

no_cvar_ptr:
	return 0;

unreadable_arg:
	Error ("can't read arg \"%s\" as a %s", vec_str_args[0].c_str(), type_name);
	return 0;
}

// returns 1 if the command is recognized, independent on an error state
bool Cvars_ParseCommandLine (vec_str_t& vec_str_line)
{
	int i_r;
	vec_cvar_t::iterator it;
	vec_cvar_t::iterator it_end;
	string* name;

	if (! vec_str_line.size())
		return 0;
	it = vec_cvars.begin();
	it_end = vec_cvars.end();
	name = &vec_str_line[0];

// loop all known vars
loop:
	if (it == it_end)
		return 0;
	i_r = name->compare (it->name);
	if (! i_r)
	{
		vec_str_t vec_str_args;

		vec_str_args.assign(vec_str_line.begin() + 1, vec_str_line.end());
		Cvars_ParseCommand (*it, vec_str_args);
		return 1;
	}
	it++;
	goto loop;
}

