// by Alexander Ilyin

#include "screen.h"
#include <windows.h>
#include <ddraw.h>
//#include "hot_keys.h"
#include <string>
using namespace std;
#include "error.h"
//#include "main.h"
#include "port.h"
// <ios> on MSVC 6.0 disables it
//#pragma warning(default:4018)
#include "notify.h"


// DirectDraw object
LPDIRECTDRAW lp_dd;
bool screen_error = 0;
#define RESOLUTION_RATES_MAX 100
forced_resolution_refreshrate_t forced_resolution_refreshrates[RESOLUTION_RATES_MAX];
uint forced_resolution_refreshrates_num = 0;
// to hold values, until timer will hit
static forced_resolution_refreshrate_t screen_setting = {0};
static bool screen_is_setting = 0;
// +++ config +++
//bool screen_preserve_cursor = 1;
bool screen_force_32bit = 0;
bool screen_double_switch_hack = 0;
// --- config ---


void GetSystemTimeAsUint64 (uint64* time)
{
	GetSystemTimeAsFileTime ((FILETIME*)time);
}

bool Screen_Mode_IsSupported (uint width, uint height, uint bpp, uint refresh)
{
	int ix;
	DEVMODE devmode;
	BOOL B_result;

// 0 initiates cache filling
	ix = 0;
// go until 0 will be returned
loop:
	B_result = EnumDisplaySettings (0, ix, &devmode);
	if (! B_result)
		return 0;
	if (devmode.dmPelsWidth != width)
		goto contnue_loop;
	if (devmode.dmPelsHeight != height)
		goto contnue_loop;
	if (devmode.dmBitsPerPel != bpp)
		goto contnue_loop;
// default refreshrate is set by specifying 0, so can be omited this way
	if (refresh)
		if (devmode.dmDisplayFrequency != refresh)
			goto contnue_loop;
// found
	return 1;
contnue_loop:
	ix++;
	goto loop;
}

bool Screen_Mode_IsSupported (screen_resolution_t* screen_resolution)
{
	return Screen_Mode_IsSupported (screen_resolution->width, screen_resolution->height, screen_resolution->bpp, screen_resolution->refresh);
}

// +++ rates list +++

bool Screen_ModeRefresh_Find (int* ix, int width, int height, int bpp, int* refresh)
{
	uint i;
	forced_resolution_refreshrate_t* refreshrate;

	for (i = 0; i < forced_resolution_refreshrates_num; i++)
	{
		refreshrate = &forced_resolution_refreshrates[i];
		if (refreshrate->width != width)
			continue;
		if (refreshrate->height != height)
			continue;
		if (refreshrate->bpp != bpp)
			continue;
		if (ix)
			*ix = i;
		if (refresh)
			*refresh = refreshrate->refresh;
		return 1;
	}
	return 0;
}

void Screen_ModeRefresh_Add (int width, int height, int bpp, int refresh)
{
	forced_resolution_refreshrate_t* refreshrate;
	int ix;
	int rate2;
	bool result;

	result = Screen_ModeRefresh_Find (&ix, width, height, bpp, &rate2);
	if (result)
//		Error ("Screen_ModeRefresh_Add () :  resolution refresh rate for mode %dx%dx%d is already registered; found at index %d, refresh %d", width, height, bpp, ix, rate2);
		Notify_Errorf ("Screen_ModeRefresh_Add () :  resolution refresh rate for mode %dx%dx%d is already registered; found at index %d, refresh %d", width, height, bpp, ix, rate2);

// todo :  check, that the mode would exist
// I meant display adapter perhaps; e-ehh, poor comments :/
	result = Screen_Mode_IsSupported (width, height, bpp, refresh);
	if (! result)
	{
// todo :  here also do not throw mbox, but log it instead and exclamate
//		Error ("Screen_ModeRefresh_Add () :  screen mode %dx%dx%dx%dHz is not supported", width, height, bpp, refresh);
		ErrorReport ("Screen_ModeRefresh_Add () :  screen mode %dx%dx%dx%dHz is not supported", width, height, bpp, refresh);
		return;
	}

	if (forced_resolution_refreshrates_num >= RESOLUTION_RATES_MAX)
		Error ("Screen_ModeRefresh_Add () :  max %u resolution refreshrates reached", RESOLUTION_RATES_MAX);
	refreshrate = &forced_resolution_refreshrates[forced_resolution_refreshrates_num];
	refreshrate->width = width;
	refreshrate->height = height;
	refreshrate->bpp = bpp;
	refreshrate->refresh = refresh;
	forced_resolution_refreshrates_num++;
}

bool Screen_WholeStruct_GetSize (uint* size, bool refreshlock)
{
	Assert(size);
	*size = sizeof(port_begin_t);
//	if (refreshlock)
		*size += sizeof(forced_resolution_refreshrate_t) * forced_resolution_refreshrates_num;
	*size += sizeof(port_end_t);
	return 1;
}

void Screen_WholeStruct_Pack (void* buffer, bool refreshlock)
{
	byte* ptr;
//	forced_resolution_refreshrate_t* refreshrate_ptr;
	port_begin_t* port_begin;
	port_end_t* port_end;
	uint video_structures_size;

	ptr = (byte*)buffer;
	port_begin = (port_begin_t*)ptr;
	port_begin->id0 = DLL_PORT_ID_0;
	port_begin->id1 = DLL_PORT_ID_1;
	port_begin->version = PORT_VERSION;
// refreshlock disabling by giving no modes
// no, bad, better give flags
//	port_begin->config_settings.flags = 0;
//	if (refreshlock)
//		port_begin->config_settings.flags |= PORT_FLAG_REFRESHLOCK;
	port_begin->config_settings = g_config_settings;
	port_begin->video_mode_struct_size = sizeof(forced_resolution_refreshrate_t);
	port_begin->video_modes_num = forced_resolution_refreshrates_num;
	port_begin->id2 = DLL_PORT_ID_2;
	port_begin->id3 = DLL_PORT_ID_3;
	ptr += sizeof(port_begin_t);

//	refreshrate_ptr = (forced_resolution_refreshrate_t*)ptr;
	video_structures_size = sizeof(forced_resolution_refreshrate_t) * forced_resolution_refreshrates_num;
//	memcpy (refreshrate_ptr);
	memcpy (ptr, forced_resolution_refreshrates, video_structures_size);
//	(*(byte**)&ptr) += size;
	ptr += video_structures_size;

	port_end = (port_end_t*)ptr;
	port_end->id4 = DLL_PORT_ID_4;
	port_end->id5 = DLL_PORT_ID_5;
	port_end->id6 = DLL_PORT_ID_6;
	port_end->id7 = DLL_PORT_ID_7;
// logical end :)
	ptr += sizeof(port_end_t);
}

// --- rates list ---

// +++ config +++

bool Screen_ParseConfigMode (string val0, string val1, string val2, int refresh)
{
	int width;
	int height;
	int bpp;
	int i_result;
	const char* arg;

	arg = val0.c_str ();
	i_result = sscanf (arg, "%d", &width);
	if (i_result != 1)
		Error ("resolution width argument \"%s\" cannot be converted to integer", arg);
	arg = val1.c_str ();
	i_result = sscanf (arg, "%d", &height);
	if (i_result != 1)
		Error ("resolution height argument \"%s\" cannot be converted to integer", arg);
	arg = val2.c_str ();
	i_result = sscanf (arg, "%d", &bpp);
	if (i_result != 1)
		Error ("resolution bpp argument \"%s\" cannot be converted to integer", arg);

	Screen_ModeRefresh_Add (width, height, bpp, refresh);

	return 1;
}

// some time will redo it to C++ :)
bool Screen_ParseString_ModesConfig (const char* cfg_str)
{
	bool result;
	int i_result;
	string str;
	string str1;
	int length;
	static int freq = 0;
	uint arg_begin[3];
	uint arg_end[3];
	int end;
	const char* arg;

	str = cfg_str;
	length = str.length ();
	if (str[0] == '[')
		goto parse_freq;

	arg_begin[0] = 0;
	arg_end[0] = str.find (' ');
	if (arg_end[0] == string::npos)
		Error ("only one argument in string \"%s\"", cfg_str);
	arg_begin[1] = arg_end[0] + 1;
	arg_end[1] = str.find (' ', arg_begin[1]);
	if (arg_end[1] == string::npos)
		Error ("only two arguments in string \"%s\"", cfg_str);
	arg_begin[2] = arg_end[1] + 1;
	arg_end[2] = str.find (' ', arg_begin[2]);
	if (arg_end[2] != string::npos)
		Error ("more than three arguments in string \"%s\"", cfg_str);
	arg_end[2] = length;

	result = Screen_ParseConfigMode
		(
		str.substr (arg_begin[0], arg_end[0] - arg_begin[0]),
		str.substr (arg_begin[1], arg_end[1] - arg_begin[1]),
		str.substr (arg_begin[2], arg_end[2] - arg_begin[2]),
		freq
		);

	return 1;

parse_freq:
	if (length < 3)
		Error ("no integer argument in brackets in string \"%s\"", cfg_str);
	end = length - 1;
	if (str[end] != ']')
		Error ("no closing bracket at the end of string \"%s\"", cfg_str);
	str1 = str.substr (1, end - 1);
	arg = str1.c_str ();
	i_result = sscanf (arg, "%d", &freq);
	if (i_result != 1)
		Error ("frequence \"%s\" cannot be converted to integer", arg);

	return 1;
}

// --- config ---

void RecreateGammaRamp (HDC hdc)
{
	ushort gamma_ramp[3][256];
	BOOL result;
	int j;

	for (j = 0; j < 256; j++)
	{
		short val = j << 8;

		gamma_ramp[0][j] = val;
		gamma_ramp[1][j] = val;
		gamma_ramp[2][j] = val;
	}

	result = SetDeviceGammaRamp (hdc, gamma_ramp);
}

//typedef DirectDrawCreate DirectDrawCreate_ptr;
typedef HRESULT (WINAPI* DirectDrawCreate_t) (GUID FAR* lpGUID, LPDIRECTDRAW FAR* lplpDD, IUnknown FAR* pUnkOuter);
DirectDrawCreate_t DirectDrawCreate_ptr = 0;

//void Screen_Restore ()
void Screen_Restore (void*)
{
	int i_r;
	HDC h_dc;

	{
// switch back to the desktop
// *refresh rate also can be set here; DEVMODE mode; mode.dmDisplayFrequency = 70; mode.dmFields = DM_DISPLAYFREQUENCY;
// refreshlock will ignore this, because NULL is instead of 'DEVMODE'
		ChangeDisplaySettings (0, 0);
//		PlaySound ("MouseClick", 0, 0);
// this works good, it does remove the left top onscreen bug, but only from a host thread, not from here
//		SetWindowPos (GetActiveWindow (), 0, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOZORDER);
// this does not work, even if the top window is the buggied one
//		SetWindowPos (GetTopWindow (0), 0, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOZORDER);
// show mouse pointer
		ShowCursor (TRUE);
	}

	h_dc = GetDC (0);

	if (h_dc)
	{
		RecreateGammaRamp (h_dc);
		i_r = ReleaseDC (0, h_dc);
	}

HMODULE h_ddraw;

//	h_ddraw = GetModuleHandle (TEXT("DDRAW"));
	h_ddraw = LoadLibrary (TEXT("DDRAW"));

	if (h_ddraw)
		*(FARPROC*)&DirectDrawCreate_ptr = GetProcAddress (h_ddraw, "DirectDrawCreate");

// "Set brains" to video card
	if (DirectDrawCreate_ptr)
	{
		DirectDrawCreate_ptr (NULL, &lp_dd, NULL);
		lp_dd->Release ();
	}
}

#if USE_WNDPROC_MESSAGES
// return 0, when need to request the exit
// *for the 'REFRESHLOCK_CHANGE_DELAYED', the 'is_screen_changed' is not optional
bool Screen_OnDisplayChange (int width, int height, int bpp, bool* is_screen_changed)
{
	HDC hdc;
	bool result;
	int i_result;
	static uint64 time_last = 0;
	uint64 time;
	static int fast_switches = 0;
//	bool needs_override = 0;
	int override_rate = 60;

	Assert(is_screen_changed);
	if (is_screen_changed)
		*is_screen_changed = 0;
// do not process anymore; just wait exit
	if (screen_error)
// do not request exit again
		return 1;
// do not check, if must soon change, or is changing
	if (screen_is_setting)
		return 1;
// gets desktop dc ?
	hdc = GetDC (0);
	if (! hdc)
		return 1;
// *not works with 0 hdc
	i_result = GetDeviceCaps (hdc, VREFRESH);
// 0 and 1 means default
	if (i_result < 2)
		return 1;

// +++ check modes +++
	result = Screen_ModeRefresh_Find (0, width, height, bpp, &override_rate);
	if (! result)
		return 1;
// test
//	override_rate = 85;
	if (i_result == override_rate)
		return 1;
// --- check modes ---

//	if (! needs_override)
//		return 1;

// +++ protect against fast switching +++
	GetSystemTimeAsUint64 (&time);
//	if (time - time_last >= TI_SECOND)
// test
	if (time - time_last >= TI_SECOND * 5)
	{
		fast_switches = 0;
		goto change_rate;
	}
	fast_switches++;
#define FAST_SWITCHES_MAX 3
	if (fast_switches < FAST_SWITCHES_MAX)
		goto change_rate;
	screen_error = 1;
// hmmm, is this good to request mbox from callback subr. (?) and what if request from win hook (?)
	ErrorReport ("fast screen modes switch %u times; exiting", FAST_SWITCHES_MAX);
//	goto finish;
// request exit
	return 0;
// --- protect against fast switching ---

change_rate:
	if (screen_force_32bit)
		bpp = 32;
// test
//	ChangeDisplaySettings (0, 0);
#if REFRESHLOCK_CHANGE_IMMEDIATE
	DEVMODE devmode;

	memset (&devmode, 0, sizeof(devmode));
	devmode.dmSize = sizeof(devmode);
	devmode.dmDisplayFrequency = override_rate;
// something is improper, if just request to change frequence;
// some problem appears, if requested 16bit mode, while desktop is 32bit;
// so must request to change all parameters
#if 0
	devmode.dmFields = DM_DISPLAYFREQUENCY;
#else
	devmode.dmPelsWidth = width;
	devmode.dmPelsHeight = height;
	devmode.dmBitsPerPel = bpp;
	devmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;
#endif
// *cursor stuff belongs only at this program
//	bool is_cursor;
//	is_cursor = !! GetCursor ();
	ChangeDisplaySettings (&devmode, 0);
//	if (screen_preserve_cursor)
//		ShowCursor (is_cursor);
#endif
#if REFRESHLOCK_CHANGE_DELAYED
	screen_setting.width = width;
	screen_setting.height = height;
	screen_setting.bpp = bpp;
	screen_setting.refresh = override_rate;
	screen_is_setting = 1;
#endif
	if (is_screen_changed)
		*is_screen_changed = 1;

//finish:
	time_last = time;
	return 1;
}

#if REFRESHLOCK_CHANGE_DELAYED
void Screen_OnTimer ()
{
	DEVMODE devmode;

	if (! screen_is_setting)
		return;
//	screen_is_setting = 0;
	memset (&devmode, 0, sizeof(devmode));
	devmode.dmSize = sizeof(devmode);
	devmode.dmPelsWidth = screen_setting.width;
	devmode.dmPelsHeight = screen_setting.height;
	devmode.dmBitsPerPel = screen_setting.bpp;
	devmode.dmDisplayFrequency = screen_setting.refresh;
	devmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;
//	bool is_cursor;
// both method gives 1 always :/
// first method
//	is_cursor = !! GetCursor ();
// second method
//	ShowCursor (1);
//	is_cursor = ShowCursor (0) >= 0;
// horrible hack, but it works
// if once change the screen resolution immediately in the 'WM_DISPLAYCHANGE',
// then the program screen remains in left top (640x480 perhaps) part of screen.
// and if not use this additional request, then the screen also same bugged, if change 50ms after 'WM_DISPLAYCHANGE'
// and if change 200ms after 'WM_DISPLAYCHANGE', then the screen is ok, but Half-Life 1.1.1.0 buggied cursor appears
// but not helps, if second resolution will be the same
#if DUAL_SWITCH_HACK
	if (screen_double_switch_hack)
	{
// no, this not works for 32bit, if desktop is 32bit
#if 0
		ChangeDisplaySettings (0, 0);
#endif
// no, this not works for 16bit
#if 0
// now it is even more crappy hack :) (would be good to make that would take existing resolution)
		devmode.dmBitsPerPel = screen_setting.bpp == 32 ? 16 : 32;
		ChangeDisplaySettings (&devmode, 0);
		devmode.dmBitsPerPel = screen_setting.bpp;
#endif
// twice horrible hack :)
		if (screen_setting.bpp == 32)
		{
			devmode.dmBitsPerPel = 16;
			ChangeDisplaySettings (&devmode, 0);
			devmode.dmBitsPerPel = screen_setting.bpp;
		}
		else
			ChangeDisplaySettings (0, 0);
// this sleep seems does not helps in anything
// 500ms bugs HL, 200ms not bugs
//		Sleep (50);
	}
#endif
	ChangeDisplaySettings (&devmode, 0);
//	if (screen_preserve_cursor)
//		ShowCursor (is_cursor);
// test
//		ShowCursor (0);
// allow resolution check only now
	screen_is_setting = 0;
}
#endif
#endif

// +++ hotkeys +++

void Screen_Hotkey_ChangeScreenResolution (void* user_data)
{
	screen_resolution_t* screen_resolution;
	DEVMODE devmode;
	LONG L_result;

	screen_resolution = (screen_resolution_t*)user_data;

	memset (&devmode, 0, sizeof(devmode));
	devmode.dmSize = sizeof(devmode);
	devmode.dmPelsWidth = screen_resolution->width;
	devmode.dmPelsHeight = screen_resolution->height;
	devmode.dmBitsPerPel = screen_resolution->bpp;
	devmode.dmDisplayFrequency = screen_resolution->refresh;
	devmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;

	L_result = ChangeDisplaySettings (&devmode, 0);
	if (L_result != DISP_CHANGE_SUCCESSFUL)
	{
		MessageBeep (MB_ICONHAND);
	}
}

// --- hotkeys ---

