// by Alexander Ilyin
// in MSVC 2005 MSDN the system tray is named Status Area
// and the whole rectangle below desktop is named Shell Tray

#include "trayicon.h"
#include <stdio.h>
#include "error.h"
#include "wnd_console.h"
#include "notify.h"
#include "trayicon_utils.h"


NOTIFYICONDATA nid;

typedef struct
{
	byte b;
	byte g;
	byte r;
	byte a;
} bgra8_t;

#define ICON_WIDTH 16
#define ICON_HEIGHT 16
//#define ICON_WIDTH 32
//#define ICON_HEIGHT 32
#define ICON_PIXELS_NUM (ICON_WIDTH * ICON_HEIGHT)
#define ICON_MASK_BYTES_NUM (ICON_WIDTH * ICON_HEIGHT / 8)
#define ICON_ID 100

bool tray_icon__initialized = 0;

static bgra8_t icon_bitmap[ICON_PIXELS_NUM];
static byte icon_bitmap_mask[ICON_MASK_BYTES_NUM];

HBITMAP tray_icon__hbitmap = 0;
HBITMAP tray_icon__hbitmap_mask = 0;
//HICON tray_icon__hicon = 0;

// +++ config +++
int graph_lines_num = 3;
// *fill colour values with some obvious nonsense; it must be set properly in the config
ulong graph_background_color = 0x80000000;
// green
ulong graph_line_color_0 = 0x80808080;
// yellow
ulong graph_line_color_1 = 0x80808080;
// red
ulong graph_line_color_2 = 0x80808080;
// purple
ulong graph_line_color_3 = 0x80808080;
// blue
ulong graph_line_color_4 = 0x80808080;
ulong graph_line_functions[5] = {0};
bool trayicon_patch = 1;
// --- config ---

// 0 none, 1 green, 2 yellow, 3 red
uint tray_icon_notification_level = 0;


// no win messages on "explorer.exe" quit
// only message 0xc07c (it is not 'WM_APP | 0x4000 | WM_STYLECHANGING', it is registered message) is on "explorer.exe" start
bool TrayIcon_Icon_Add (HWND hwnd, HICON icon)
{
	bool result;
	BOOL B_result;

	nid.cbSize = sizeof(NOTIFYICONDATA);
	nid.hWnd = hwnd;
// Application-defined
// by Windows Embedded CE docs - values from 0 to 12 are reserved
//	nid.uID = 1;
	nid.uID = ICON_ID;
	nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
//	nid.uFlags = NIF_ICON;
	nid.uCallbackMessage = WM_SYSTRAY_ICON;
	nid.hIcon = icon;
	sprintf (nid.szTip, "Info tray icon start");
	B_result = Shell_NotifyIcon (NIM_ADD, &nid);

	result = !! B_result;

	return result;
}

void TrayIcon_Bitmaps_Delete ();

bool TrayIcon_Bitmaps_Create (HDC hdc)
{
	Assert(hdc);
#if 0
	if (! hdc)
// no device context handle is given, get by ownself
	{
		hdc = GetDC (hwnd);
		if (! hdc)
			return 0;
	}
#endif

// 'SetDIBits ()' is a little bit complicated, and because of that I wasn't wished to use it (CreateDIBitmap () exist also)
// 24 bits 'CreateBitmap ()' worked ok on Win98, but gives black icon on WinXP (on both, 16 and 32 bits desktop),
// and all other cases, if the bpp isn't the same as the device context have
// 'CreateBitmapIndirect ()' is the same as 'CreateBitmap ()'
// so it not uses device-independent bitmap (DIB) anymore
	tray_icon__hbitmap = CreateCompatibleBitmap (hdc, ICON_WIDTH, ICON_HEIGHT);
	if (! tray_icon__hbitmap)
		return 0;
	tray_icon__hbitmap_mask = CreateBitmap (ICON_WIDTH, ICON_HEIGHT, 1, 1, icon_bitmap_mask);
	if (! tray_icon__hbitmap_mask)
	{
		TrayIcon_Bitmaps_Delete ();
		return 0;
	}
	return 1;
}

void TrayIcon_Bitmaps_Delete ()
{
	BOOL B_r;

	if (tray_icon__hbitmap)
	{
		B_r = DeleteObject (tray_icon__hbitmap);
		tray_icon__hbitmap = 0;
	}
	if (tray_icon__hbitmap_mask)
	{
		B_r = DeleteObject (tray_icon__hbitmap_mask);
		tray_icon__hbitmap_mask = 0;
	}
}

typedef struct BITMAPINFO_s
{
	BITMAPINFOHEADER bmiHeader;
	ulong color_masks[4];
} BITMAPINFO_t;

bool TrayIcon_DrawIntoBitmap (HDC hdc, HBITMAP hbitmap)
{
	int i_result;
//	BITMAPINFO bitmapinfo;
	BITMAPINFO_t bitmapinfo;
//	BITMAPINFOHEADER bitmapinfoheader;

//	memset (&bitmapinfoheader, 0, sizeof(bitmapinfoheader));
//	bitmapinfoheader.biSize = sizeof(bitmapinfoheader);
	memset (&bitmapinfo.bmiHeader, 0, sizeof(bitmapinfo.bmiHeader));
	bitmapinfo.bmiHeader.biSize = sizeof(bitmapinfo.bmiHeader);
	bitmapinfo.bmiHeader.biWidth = ICON_WIDTH;
	bitmapinfo.bmiHeader.biHeight = -ICON_HEIGHT;
	bitmapinfo.bmiHeader.biPlanes = 1;
	bitmapinfo.bmiHeader.biBitCount = 32;
//	bitmapinfo.bmiHeader.biCompression = BI_RGB;
	bitmapinfo.bmiHeader.biCompression = BI_BITFIELDS;
	bitmapinfo.bmiHeader.biSizeImage = 0;
//	bitmapinfo.bmiHeader = ;

#if 0
	bitmapinfo.bmiColors[0].rgbRed = 0;
	bitmapinfo.bmiColors[0].rgbGreen = 0;
	bitmapinfo.bmiColors[0].rgbBlue = 0;
	bitmapinfo.bmiColors[0].rgbReserved = 0;
//	bitmapinfo.bmiColors[0] = ;
#endif

// red
	bitmapinfo.color_masks[0] = 0x00ff0000;
// green
	bitmapinfo.color_masks[1] = 0x0000ff00;
// blue
	bitmapinfo.color_masks[2] = 0x000000ff;
// not in use :(
	bitmapinfo.color_masks[3] = 0xff000000;

// and this also is why 'SelectObject ()' isn't appearing
//	i_result = SetDIBits (hdc, hbitmap, 0, ICON_HEIGHT, icon_bitmap, &bitmapinfo, DIB_RGB_COLORS);
	i_result = SetDIBits (hdc, hbitmap, 0, ICON_HEIGHT, icon_bitmap, (BITMAPINFO*)&bitmapinfo, DIB_RGB_COLORS);

	return !! i_result;
}

// its return value is necessary in the beginning of 'WinMain ()', so it would know what to do further, so don't make it 'void'
// but now I think that it would be more proper, if it will return 0 in all failure cases, except of 'TrayIcon_Icon_Add ()' failure,
// which fails, if systray isn't created yet, what is normal in some cases
// no, I need more complex return value :
// 0 - some fail, which shouldn't normally occur,
// 1 - everything is good, but icon add has failed,
// 2 - everything succeed
// no ! (hehe) as did earlier :  return 0 on an unexpected error, 1 otherwise
//bool TrayIcon_Init (HWND hwnd)
//uint TrayIcon_Init (HWND hwnd)
bool TrayIcon_Init (HWND hwnd, bool* icon_add_failure)
{
	bool r;
//	bool initialized = 0;
	bool no_non_systray_failures = 0;
	int i_result;
	BOOL B_result;
//	HBITMAP h_bitmap;
//	HBITMAP h_bitmap_mask;
	HICON hicon;
	ICONINFO iconinfo;
	int i;
	HDC hdc;

	tray_icon__initialized = 0;

	for (i = 0; i < ICON_PIXELS_NUM; i++)
	{
		icon_bitmap[i].r = 0xff;
		icon_bitmap[i].g = 0x80;
		icon_bitmap[i].b = 0x40;
		icon_bitmap[i].a = 0xff;
	}

// if rgba, then mask is not in use, but without mask the 'CreateIconIndirect ()' doesn't work
#if 0
	for (i = 0; i < ICON_MASK_BYTES_NUM; i++)
	{
		icon_bitmap_mask[i] = 0x00;
	}
#endif

	hdc = GetDC (hwnd);
	if (! hdc)
//		goto finish;
// *'icon_add_failure' set is skipped in this case, but that isn't matters for that case
		return 0;
//		return;

	r = TrayIcon_Bitmaps_Create (hdc);
	if (! r)
		goto finish;
	r = TrayIcon_DrawIntoBitmap (hdc, tray_icon__hbitmap);
	if (! r)
//		goto delete_bitmap;
		goto finish;

	iconinfo.fIcon = 1;
	iconinfo.xHotspot = 0; 
	iconinfo.yHotspot = 0; 
	iconinfo.hbmMask = tray_icon__hbitmap_mask;
	iconinfo.hbmColor = tray_icon__hbitmap; 
// it copies bitmaps; no need to remain it
	hicon = CreateIconIndirect (&iconinfo);
	if (! hicon)
//		goto delete_bitmap_mask;
//		goto delete_bitmap;
		goto finish;

// return 1, if icon add will fail
//// maybe will be good even for WM_TASKBARCREATED
	no_non_systray_failures = 1;
	r = TrayIcon_Icon_Add (hwnd, hicon);
//	initialized = result;
	tray_icon__initialized = r;

	B_result = DestroyIcon (hicon);

// less significant, secondary stuff
	if (trayicon_patch)
		TrayIconUtils_Patch (hwnd, ICON_ID);

finish:
	if (! tray_icon__initialized)
// free allocated objects
		TrayIcon_Close ();

//release_dc:
	i_result = ReleaseDC (hwnd, hdc);

	if (icon_add_failure)
		*icon_add_failure = ! tray_icon__initialized;

// yes, the change of kind of the return value; will be a bit messing, but tend your mind to understand it :)
//	return initialized;
	return no_non_systray_failures;
#if 0
	if (tray_icon__initialized)
		return 2;
	else
		return no_non_systray_failures;
#endif
}

//bool TrayIcon_Close ()
void TrayIcon_Close ()
{
//	bool result;
	BOOL B_result;

	if (tray_icon__initialized)
	{
		tray_icon__initialized = 0;
// should fail, in case of 'WM_TASKBARCREATED'
		B_result = Shell_NotifyIcon (NIM_DELETE, &nid);
	}

//	result = !! B_result;

//	B_result = DestroyIcon (tray_icon__hicon);

	TrayIcon_Bitmaps_Delete ();

//	return result;
}

// compatible bitmaps depend on the screen context; have to update them in some cases
void TrayIcon_ResolutionChanged (HWND hwnd)
{
	int i_r;
	HDC hdc;

	TrayIcon_Bitmaps_Delete ();
	hdc = GetDC (hwnd);
	if (! hdc)
		return;
	TrayIcon_Bitmaps_Create (hdc);
	i_r = ReleaseDC (hwnd, hdc);
}

//bool TrayIcon_LineSetFunction (uint line_ix, string& func_name)
bool TrayIcon_LineSetFunction (uint line_ix, const char* func_name)
{
//	uint i;
	int i_result;

	Assert(line_ix < NUMOF(graph_line_functions));
	Assert(TRAYICON_LINE_FUNCS_NUM == trayicon_line_funcs_num);

	FORI0(TRAYICON_LINE_FUNCS_NUM)
	{
		i_result = _stricmp (func_name, trayicon_line_funcs_names[i]);
		if (i_result)
			continue;
		graph_line_functions[line_ix] = i;
		return 1;
	}
	Error ("TrayIcon_LineSetFunction () :  icon line %u; function \"%s\" is not found", line_ix, func_name);
	return 0;
}

// schemes are :
// 1+4+1+4+1+4+1 = 4*3 + 1*4 = 16
// 1+3+1+3+1+3+1+3 (not good) = 4*4 = 16
// 1+2+1+2+1+2+1+2+1+2+1 = 2*5 + 1*6 = 16

void TrayIcon_DrawLine_Three (ulong* pixels_line, int height, int l0, int l1, int l2)
{
// line 0
	if (height < l0)
	{
		pixels_line[1] = graph_line_color_0;
		pixels_line[2] = graph_line_color_0;
		pixels_line[3] = graph_line_color_0;
		pixels_line[4] = graph_line_color_0;
	}
// line 1
	if (height < l1)
	{
		pixels_line[6] = graph_line_color_1;
		pixels_line[7] = graph_line_color_1;
		pixels_line[8] = graph_line_color_1;
		pixels_line[9] = graph_line_color_1;
	}
// line 2
	if (height < l2)
	{
		pixels_line[11] = graph_line_color_2;
		pixels_line[12] = graph_line_color_2;
		pixels_line[13] = graph_line_color_2;
		pixels_line[14] = graph_line_color_2;
	}
}

void TrayIcon_DrawLine_Five (ulong* pixels_line, int height, int l0, int l1, int l2, int l3, int l4)
{
// line 0
	if (height < l0)
	{
		pixels_line[1] = graph_line_color_0;
		pixels_line[2] = graph_line_color_0;
	}
// line 1
	if (height < l1)
	{
		pixels_line[4] = graph_line_color_1;
		pixels_line[5] = graph_line_color_1;
	}
// line 2
	if (height < l2)
	{
		pixels_line[7] = graph_line_color_2;
		pixels_line[8] = graph_line_color_2;
	}
// line 3
	if (height < l3)
	{
		pixels_line[10] = graph_line_color_3;
		pixels_line[11] = graph_line_color_3;
	}
// line 4
	if (height < l4)
	{
		pixels_line[13] = graph_line_color_4;
		pixels_line[14] = graph_line_color_4;
	}
}

//bool TrayIcon_Update_ThreeGraphLines (HWND hwnd, char* tooltip, fix line0, fix line1, fix line2)
bool TrayIcon_Update_GraphLines (HWND hwnd, char* tooltip, fix line0, fix line1, fix line2, fix line3, fix line4)
{
	bool r;
	bool updated = 0;
	int i_r;
	BOOL B_r;
	HICON hicon;
	HDC hdc;
	ICONINFO iconinfo;
	int i;
	int l0;
	int l1;
	int l2;
	int l3;
	int l4;

	updated = 0;
	hdc = 0;

// give by 1 line for the top and bottom border
#define GRAPH_HEIGHT (ICON_HEIGHT - 2)

	for (i = 0; i < ICON_PIXELS_NUM; i++)
	{
		*(ulong*)&icon_bitmap[i] = graph_background_color;
	}

// events notification
	ulong color_notification = 0;
	switch (tray_icon_notification_level)
	{
	case 0:
		break;
	case 1:
		color_notification = 0xff00ff00;
		break;
	case 2:
		color_notification = 0xffffff00;
		break;
	case 3:
		color_notification = 0xffff0000;
		break;
	default:
		Error ("TrayIcon_Update_GraphLines () :  unknown notification level %u", tray_icon_notification_level);
	}
	if (tray_icon_notification_level)
	{
		for (i = 0; i < ICON_WIDTH; i++)
		{
//			*(ulong*)&icon_bitmap[i + ICON_WIDTH * 0] = color_notification;
//			*(ulong*)&icon_bitmap[i + ICON_WIDTH * 1] = color_notification;
//			*(ulong*)&icon_bitmap[i + ICON_WIDTH * 2] = color_notification;
			*(ulong*)&icon_bitmap[i + ICON_WIDTH * (ICON_HEIGHT - 1)] = color_notification;
			*(ulong*)&icon_bitmap[i + ICON_WIDTH * (ICON_HEIGHT - 2)] = color_notification;
			*(ulong*)&icon_bitmap[i + ICON_WIDTH * (ICON_HEIGHT - 3)] = color_notification;
		}
	}

	if (ICON_WIDTH < 16)
		goto get_dc;

// scale lines vertically to graph height
	l0 = line0 * GRAPH_HEIGHT;
	l0 = (l0 + F0_5) / F1_0;
	l1 = line1 * GRAPH_HEIGHT;
	l1 = (l1 + F0_5) / F1_0;
	l2 = line2 * GRAPH_HEIGHT;
	l2 = (l2 + F0_5) / F1_0;
	l3 = line3 * GRAPH_HEIGHT;
	l3 = (l3 + F0_5) / F1_0;
	l4 = line4 * GRAPH_HEIGHT;
	l4 = (l4 + F0_5) / F1_0;

	for (i = 0; i < GRAPH_HEIGHT; i++)
	{
		ulong* pixels_line = (ulong*)&icon_bitmap[(i + 1) * ICON_WIDTH];
		int height = GRAPH_HEIGHT - i - 1;

		switch (graph_lines_num)
		{
		case 3:
			TrayIcon_DrawLine_Three (pixels_line, height, l0, l1, l2);
			break;
		case 5:
			TrayIcon_DrawLine_Five (pixels_line, height, l0, l1, l2, l3, l4);
			break;
		default:
			Error ("wrong graph_lines_num value %d", graph_lines_num);
		}
	}

get_dc:
	hdc = GetDC (hwnd);
	if (! hdc)
		goto finish;
// this normally shouldn't occur, if everything around is working correct
	if (! tray_icon__hbitmap)
	{
		r = TrayIcon_Bitmaps_Create (hdc);
//		WndConsole_Printf ("TrayIcon_Update_GraphLines () :  device context compatible bitmaps weren't allocated; new allocation result is %u", r);
		Notify_Printf ("TrayIcon_Update_GraphLines () :  device context compatible bitmaps weren't allocated; new allocation result is %u", r);
	}
	r = TrayIcon_DrawIntoBitmap (hdc, tray_icon__hbitmap);
//	i_r = ReleaseDC (hwnd, hdc);
	if (! r)
		goto finish;

	iconinfo.fIcon = 1;
	iconinfo.xHotspot = 0; 
	iconinfo.yHotspot = 0; 
	iconinfo.hbmMask = tray_icon__hbitmap_mask;
	iconinfo.hbmColor = tray_icon__hbitmap; 
// I think it is bad, to create icon every time anew; maybe rewrite it later
	hicon = CreateIconIndirect (&iconinfo);

	nid.uFlags = NIF_ICON | NIF_TIP;
//	nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	nid.hIcon = hicon;
	sprintf (nid.szTip, "%.63s", tooltip);
	nid.szTip[63] = 0;

	B_r = Shell_NotifyIcon (NIM_MODIFY, &nid);

	updated = !! B_r;
// last error 0 is returned, if the function has been failed :/
// *after 'Shell_NotifyIcon ()' fail, it still returns 0 for me
#if 0
	DWORD D_r;
	D_r = GetLastError ();
	if (! updated)
		Error ("notification area icon update failed, error %u", D_r);
#endif

// since 'GetLastError ()' does not help, have to do own tests for the fail case
	if (! B_r)
// try to find out, maybe the screen dc just become incompatible
// can also be, that the notify icon id become invalid
	{
		HDC hdc_compatible;
		HBITMAP hbm_old;

		Assert(hdc);
		Assert(tray_icon__hbitmap);
		hdc_compatible = CreateCompatibleDC (hdc);
		if (! hdc_compatible)
			Error ("TrayIcon_Update_GraphLines () :  cannot create a dc, compatible to screen");
		hbm_old = (HBITMAP)SelectObject (hdc_compatible, tray_icon__hbitmap);
		if (hbm_old)
		{
			SelectObject (hdc_compatible, hbm_old);
		}
		else
// not ok
		{
			WndConsole_Print ("TrayIcon_Update_GraphLines () :  bitmap is incompatible with the device context; recreating");
			TrayIcon_Bitmaps_Delete ();
			r = TrayIcon_Bitmaps_Create (hdc);
		}
		DeleteDC (hdc_compatible);
	}

	B_r = DestroyIcon (hicon);

finish:
	if (hdc)
		i_r = ReleaseDC (hwnd, hdc);

	return updated;
}

